from walking_module import WalkingModule

import json
import math
import os
import sys
from time import sleep
from events.msg import new_pos
import roslib.packages as rospkg
import rospy
from std_msgs.msg import String, Bool
import numpy as np


class AdvancedWalkingModule(WalkingModule):

    def __init__(self, event, robot_size):

        super(AdvancedWalkingModule, self).__init__(event, robot_size)
        self.plan_path = []
        self.is_walking = False

        self.plan_orientation = {'angle': 0, 'x': 0, 'y': 0, 'distance': 0, 'offset': 0}

        self.step_counter = 0

        self.current_plan_walking_mode = "turn_right"

        self.event = event

        self.is_half_step_done_sub = rospy.Subscriber('/robotis/walking/is_full_step_done', Bool,
                                                      self.process_update_step_info)
        self.calibration_factors = None

        self.calibration_factors = self.load_walking_calibration_factors(self.robot_size)

        self.new_pos_pub = rospy.Publisher(
            '/events/advance_walking_module/new_pos_out', new_pos, queue_size=0)

        self.new_pos_sub = rospy.Subscriber(
            '/events/advance_walking_module/new_pos_in', new_pos, self.process_new_pos_sub)

    def process_new_pos_publish(self):
        pos = new_pos()
        pos.angle = math.radians(self.plan_orientation['angle'])
        pos.x = self.plan_orientation['x']
        pos.y = self.plan_orientation['y']
        self.new_pos_pub.publish(pos)

        print """Advanced Walking: X: %.2f, Y: %.2f, angle: %.2f, offset angle: %.2f""" % (
            self.plan_orientation['x'], self.plan_orientation['y'], self.plan_orientation['angle'],
            self.plan_orientation['offset'])

    def load_walking_calibration_factors(self, robot_size):
        calibration_file_path = os.path.join(rospkg.get_pkg_dir('events'), 'config', self.event,
                                             'walk_mappings_' + robot_size + '.json')

        try:
            with open(calibration_file_path, 'r') as fp:
                config = json.load(fp)
                config['turn_left']['rotation_angle'] = config['turn_left']['rotation_angle']
                config['turn_right']['rotation_angle'] = config['turn_right']['rotation_angle']
        except:

            config = self.create_calibration_file()
            self.save_walking_calibration_factors(robot_size, config)

        return config

    def save_walking_calibration_factors(self, robot_size, config=None):
        calibration_file_path = os.path.join(rospkg.get_pkg_dir('events'), 'config', self.event,
                                             'walk_mappings_' + robot_size + '.json')
        fp = open(calibration_file_path, 'w')
        temp = self.create_calibration_file()
        save_config = self.calibration_factors if config is None else config

        temp['forward']['distance'] = save_config['forward']['distance']
        temp['backward']['distance'] = save_config['backward']['distance']

        temp['shift_left']['distance'] = save_config['shift_left']['distance']
        temp['shift_right']['distance'] = save_config['shift_right']['distance']
        temp['turn_left']['rotation_angle'] = save_config['turn_left']['rotation_angle']
        temp['turn_right']['rotation_angle'] = save_config['turn_right']['rotation_angle']

        json.dump(save_config, fp, sort_keys=True, indent=4)
        fp.close()

    def process_update_step_info(self, msg):
        if msg.data == True:
            self.step_counter = self.step_counter + 1
            # if self.step_counter == 0:
            # self.actual_path.append({'type': self.current_actual_walking_mode, 'step':1})
            self.update_position()
            self.process_new_pos_publish()
            # print self.current_plan_walking_mode
            # print len(self.plan_path)

    def process_new_pos_sub(self, msg):
        self.plan_orientation['x'] = msg.x
        self.plan_orientation['y'] = msg.y
        self.plan_orientation['angle'] = math.degrees(msg.angle)
        self.plan_orientation['distance'] = 0
        self.plan_orientation['offset'] = 0
        self.plan_path = []
        print self.plan_orientation

    def create_calibration_file(self):
        return {

            "forward": {
                "distance": 2,
            },
            "backward": {
                "distance": -2,
            },

            "shift_left": {
                "distance": 2,
            },
            "shift_right": {
                "distance": 2,
            },

            "turn_left": {

                "rotation_angle": 5
            },
            "turn_right": {

                "rotation_angle": -5
            }
        }

    def calibration_prepare(self):
        self.step_counter = 0
        self.plan_path = []

    def degree_to_rad(self, degree):
        return degree * math.pi / 180

    def rad_to_degree(self, rad):
        return rad * 180 / math.pi

    def get_offset_degree(self):
        return self.plan_orientation['offset']

    def calibration_walking(self, data, walking_type):
        if 'turn' in walking_type:

            self.calibration_factors[walking_type]['rotation_angle'] = data / self.step_counter
        else:
            self.calibration_factors[walking_type]['distance'] = data / self.step_counter

        self.save_walking_calibration_factors(self.robot_size)

    def calcuate_position(self, position, mode):
        new_position = {
            'x': position['x'],
            'y': position['y'],
            'angle': position['angle'],
            'distance': position['distance'],
            'offset': position['offset']
        }
        if not self.current_plan_walking_mode == 'stop':
            # if 'turn' not in mode:
            #     new_position['y'] = math.sin(math.radians(position['angle'])) * self.calibration_factors[mode][
            #         'distance'] + position['y']
            #     new_position['x'] = math.cos(math.radians(position['angle'])) * self.calibration_factors[mode][
            #         'distance'] + position['x']
            #     new_position['distance'] = math.sqrt(position['y'] ** 2 + position['x'] ** 2)
            #
            # else:
            #     new_position['angle'] = (position['angle'] + self.calibration_factors[mode]['rotation_angle']) % 360
            #     new_position['offset'] = (position['offset'] + self.calibration_factors[mode]['rotation_angle']) % 360

            if 'turn' in mode:
                new_position['angle'] = (position['angle'] + self.calibration_factors[mode]['rotation_angle']) % 360
                new_position['offset'] = (position['offset'] + self.calibration_factors[mode]['rotation_angle']) % 360
                new_position['angle'] = -360+ new_position['angle'] if new_position['angle']>180 else new_position['angle']
                new_position['offset'] = -360+ new_position['offset'] if new_position['offset']>180 else new_position['offset']

            elif 'shift' in mode:
                offset = 90 if 'left' in mode else -90
                new_position['y'] = math.sin(math.radians(position['angle']+offset)) * self.calibration_factors[mode][
                    'distance'] + position['y']
                new_position['x'] = math.cos(math.radians(position['angle']+offset)) * self.calibration_factors[mode][
                    'distance'] + position['x']
                new_position['distance'] = math.sqrt(position['y'] ** 2 + position['x'] ** 2)

            else:

                new_position['y'] = math.sin(math.radians(position['angle'])) * self.calibration_factors[mode][
                    'distance'] + position['y']
                new_position['x'] = math.cos(math.radians(position['angle'])) * self.calibration_factors[mode][
                    'distance'] + position['x']
                new_position['distance'] = math.sqrt(position['y']
                    ** 2 + position['x'] ** 2)


        return new_position

    def update_position(self):
        self.plan_orientation = self.calcuate_position(self.plan_orientation, self.current_plan_walking_mode)
        self.plan_path.append({'type': self.current_plan_walking_mode, 'step': 1})

    def generate_end_points(self):
        path = self.plan_path
        orientation = {'angle': 0, 'x': 0, 'y': 0, 'distance': 0}
        points = []
        for i in path:
            orientation = self.calcuate_position(orientation, i['type'])
            points.append((orientation['x'], orientation['y']))
        return points

    def forward(self, step=None, distance=None):
        print "advance"
        self.stright_walk_helper('forward', step, distance)

    def backward(self, step=None, distance=None):
        self.stright_walk_helper('backward', step, distance)

    def shift_left(self, distance=None, step=None):
        self.stright_walk_helper('shift_left',step,distance)

    def shift_right(self, distance=None, step=None):
        self.stright_walk_helper('shift_right', step, distance)

    def stright_walk_helper(self, mode, step=None, distance=None):

        self.current_plan_walking_mode = mode
        if step is None and distance is None:
            super(AdvancedWalkingModule, self).walk(gait=mode)
        elif distance is not None:
            step = abs(round(distance / self.calibration_factors[mode]['distance']))
            print "walk %d steps" % (step)
        if step is not None:
            super(AdvancedWalkingModule, self).walk(gait=mode)
            current_step_count = self.step_counter
            while self.step_counter - current_step_count < step:
                continue
            super(AdvancedWalkingModule, self).stop()
            self.current_plan_walking_mode = 'stop'
            print "walk finish"

    def turn_left(self, angle=None, step=None):
        self.turning_helper('turn_left', angle, step)

    def turn_right(self, angle=None, step=None):
        self.turning_helper('turn_right', angle, step)

    def turning_helper(self, mode, angle=None, step=None):
        self.current_plan_walking_mode = mode
        if angle is None and step is None:
            # self.current_plan_walking_mode = mode
            super(AdvancedWalkingModule, self).walk(gait=mode)
        elif angle is not None:
            step = round(abs(angle / self.calibration_factors[mode]['rotation_angle']))
            print "turning %d steps" % (step)
        if step is not None:
            current_step_count = self.step_counter
            # super(AdvancedWalkingModule, self).walk(gait=mode)
            super(AdvancedWalkingModule, self).walk(gait=mode)
            while self.step_counter - current_step_count < step:
                continue

            super(AdvancedWalkingModule, self).stop()
            self.current_plan_walking_mode = 'stop'
            print "turning finish"

    def wait_step_finish(self, target_step):
        prev_step = self.step_counter
        while self.step_counter - prev_step < target_step:
            pass

    def get_walk_step(self, gait, angle = None, distance = None):
        if 'turn' in gait:
            return round(abs(angle / self.calibration_factors[gait]['rotation_angle']))
        else:
            return abs(round(distance / self.calibration_factors[gait]['distance']))