#!/usr/bin/env python

from time import sleep

import rospy
from std_msgs.msg import String
from sensor_msgs.msg import Imu
from advanced_walking_module import AdvancedWalkingModule
from robotis_controller_msgs.msg import JointCtrlModule

from head_control_module import HeadControlModule
from action_module import ActionModule
from walking_module import WalkingModule


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(
                Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class ControlModule(ActionModule, HeadControlModule, WalkingModule):

    __metaclass__ = Singleton

    joints = [
        'r_sho_pitch', 
        'l_sho_pitch', 
        'r_sho_roll', 
        'l_sho_roll', 
        'r_el', 
        'l_el', 
        'r_hip_yaw', 
        'l_hip_yaw', 
        'r_hip_roll', 
        'l_hip_roll', 
        'r_hip_pitch', 
        'l_hip_pitch', 
        'r_knee', 
        'l_knee', 
        'r_ank_pitch', 
        'l_ank_pitch', 
        'r_ank_roll', 
        'l_ank_roll', 
        'head_pan', 
        'head_tilt'
    ]

    def __init__(self, event, robot_size):
        super(ControlModule, self).__init__(event, robot_size)

        self.__controlling_module = {'head': None, 'body': None}

        # Initialize OP3 ROS publishers and service clients
        self.__module_control_pub = rospy.Publisher(
            '/robotis/set_joint_ctrl_modules', JointCtrlModule, queue_size=0)

        imu_sub = rospy.Subscriber('/fallen', String, self.imu_callback)
        self.fallen_status = 'STAND'


    def __enable(self, module):
        # Call service to enable module
        message = JointCtrlModule()
        for joint in ControlModule.joints:
            if HeadControlModule.head_enabled and joint.find('head') != -1:
                    message.joint_name.append(joint)
                    message.module_name.append('head_control_module')

            if WalkingModule.walk_enabled:
                if joint.find('sho') != -1 or joint.find('el') != -1:
                    message.joint_name.append(joint)
                    message.module_name.append('action_module')
                elif joint.find('head') == -1:
                    message.joint_name.append(joint)
                    message.module_name.append('walking_module')
 
    
            if module == 'action_module':
                message.joint_name.append(joint)
                message.module_name.append(module)

        
        self.__module_control_pub.publish(message)

        # Important sleep to wait for the framework to switch modules
        sleep(1)

    def action(self):
        if not ActionModule.action_enabled:

            # Stop walking before playing action
            if WalkingModule.walk_enabled:
                self.stop()

            ActionModule.action_enabled = True
            HeadControlModule.head_enabled = False
            WalkingModule.walk_enabled = False

            self.__enable(module=ActionModule.name)

        return self

    def head(self):
        if not HeadControlModule.head_enabled:

            ActionModule.action_enabled = False
            HeadControlModule.head_enabled = True

            self.__enable(module=HeadControlModule.name)

        return self

    def walking(self, gait='initial'):
        if not WalkingModule.walk_enabled:
            ActionModule.action_enabled = False
            WalkingModule.walk_enabled = True

            self.set_initial_walk(gait)
            self.__enable(module=WalkingModule.name)

        return self

    def imu_callback(self, msg):
        self.fallen_status = msg.data

    def check_fallen(self):
        return self.fallen_status


class AdvancedControlModule(ActionModule, HeadControlModule, AdvancedWalkingModule ):
    __metaclass__ = Singleton

    joints = [
        'r_sho_pitch',
        'l_sho_pitch',
        'r_sho_roll',
        'l_sho_roll',
        'r_el',
        'l_el',
        'r_hip_yaw',
        'l_hip_yaw',
        'r_hip_roll',
        'l_hip_roll',
        'r_hip_pitch',
        'l_hip_pitch',
        'r_knee',
        'l_knee',
        'r_ank_pitch',
        'l_ank_pitch',
        'r_ank_roll',
        'l_ank_roll',
        'head_pan',
        'head_tilt'
    ]

    def __init__(self, event,robot_size):
        super(AdvancedControlModule, self).__init__(event, robot_size)

        self.__controlling_module = {'head': None, 'body': None}

        # Initialize OP3 ROS publishers and service clients
        self.__module_control_pub = rospy.Publisher(
            '/robotis/set_joint_ctrl_modules', JointCtrlModule, queue_size=0)

        imu_sub = rospy.Subscriber('/fallen', String, self.imu_callback)
        self.fallen_status = 'STAND'

    def __enable(self, module):
        # Call service to enable module
        message = JointCtrlModule()
        for joint in ControlModule.joints:
            if HeadControlModule.head_enabled and joint.find('head') != -1:
                message.joint_name.append(joint)
                message.module_name.append('head_control_module')

            if WalkingModule.walk_enabled:
                if joint.find('sho') != -1 or joint.find('el') != -1:
                    message.joint_name.append(joint)
                    message.module_name.append('action_module')
                elif joint.find('head') == -1:
                    message.joint_name.append(joint)
                    message.module_name.append('walking_module')

            if module == 'action_module':
                message.joint_name.append(joint)
                message.module_name.append(module)

        self.__module_control_pub.publish(message)

        # Important sleep to wait for the framework to switch modules
        sleep(1)

    def action(self):
        if not ActionModule.action_enabled:

            # Stop walking before playing action
            if WalkingModule.walk_enabled:
                self.stop()

            ActionModule.action_enabled = True
            HeadControlModule.head_enabled = False
            WalkingModule.walk_enabled = False

            self.__enable(module=ActionModule.name)

        return self

    def head(self):
        if not HeadControlModule.head_enabled:
            ActionModule.action_enabled = False
            HeadControlModule.head_enabled = True

            self.__enable(module=HeadControlModule.name)

        return self

    def walking(self, gait='initial'):
        if not WalkingModule.walk_enabled:
            ActionModule.action_enabled = False
            WalkingModule.walk_enabled = True

            self.__enable(module=WalkingModule.name)
            self.set_initial_walk(gait)

        return self

    def imu_callback(self, msg):
        self.fallen_status = msg.data

    def check_fallen(self):
        return self.fallen_status
