#!/usr/bin/env python

import copy
import json
import math
import os
import sys
from time import sleep

import roslib.packages as rospkg
import rospy
from std_msgs.msg import String

from op3_walking_module_msgs.msg import WalkingParam


class WalkingModule(object):
    
    name = 'walking_module'
    walk_enabled = False

    DEFAULT_WALKING_PARAM_KEYS = set([
        'init_roll_offset',
        'init_pitch_offset',
        'init_yaw_offset',
        'dsp_ratio',
        'step_fb_ratio',
        'move_aim_on',
        'balance_enable',
        'balance_hip_roll_gain',
        'balance_knee_gain',
        'balance_ankle_roll_gain',
        'balance_ankle_pitch_gain',
        'y_swap_amplitude',
        'z_swap_amplitude',
        'arm_swing_gain',
        'pelvis_offset',
        'hip_pitch_offset',
        'p_gain',
        'i_gain',
        'd_gain'
    ])

    GAIT_WALKING_PARAM_KEYS = set([
        'init_x_offset',
        'init_y_offset',
        'init_z_offset',
        'x_move_amplitude',
        'y_move_amplitude',
        'z_move_amplitude',
        'angle_move_amplitude',
        'period_time',
        'hip_pitch_offset',
        'init_y_offset'
    ])

    DEGREE_WALKING_PARAM_KEYS = [
        'init_roll_offset',
        'init_pitch_offset',
        'init_yaw_offset',
        'hip_pitch_offset',
        'angle_move_amplitude',
        'pelvis_offset',
        'hip_pitch_offset',
        'init_y_offset'
    ]

    def __init__(self, event, robot_size):
        super(WalkingModule, self).__init__()
        self.initial_walk = 'initial'
        self.is_initial_walk = True
        self.robot_size = robot_size

        self.__set_walking_command_pub = rospy.Publisher(
            '/robotis/walking/command', String, queue_size=0)
        self.__set_walking_param_pub = rospy.Publisher(
            '/robotis/walking/set_params', WalkingParam, queue_size=0)

        self.default_walking_gait, self.gaits = self.__load_configuration(event, robot_size)
        self.__update_walking_gait(gait='initial')

    def __load_configuration(self, event, robot_size):
        config_file_path = os.path.join(rospkg.get_pkg_dir('events'), 'config', event, 'walking_configurations_{0}.json'.format(robot_size))

        gaits = {}

        with open(config_file_path, 'r') as fp:
            config = json.load(fp)

            default_walking_gait = dict()
            missing_gait_params = list(WalkingModule.DEFAULT_WALKING_PARAM_KEYS - set(config.keys()))

            for param in config.keys():
                if param.startswith('gait'):
                    gaits[str(param[5:])] = config[param]
                else:
                    default_walking_gait[param] = sys.maxint if param in missing_gait_params else config[param]

        return default_walking_gait, gaits

    def __set_walking_params(self, walking_param, config):
        walking_param.init_x_offset = config['init_x_offset']
        walking_param.init_y_offset = config['init_y_offset']
        walking_param.init_z_offset = config['init_z_offset']
        walking_param.init_roll_offset = config['init_roll_offset']
        walking_param.init_pitch_offset = config['init_pitch_offset']
        walking_param.init_yaw_offset = config['init_yaw_offset']
        walking_param.dsp_ratio = config['dsp_ratio']
        walking_param.step_fb_ratio = config['step_fb_ratio']
        walking_param.move_aim_on = config['move_aim_on']
        walking_param.balance_enable = config['balance_enable']
        walking_param.balance_hip_roll_gain = config['balance_hip_roll_gain']
        walking_param.balance_knee_gain = config['balance_knee_gain']
        walking_param.balance_ankle_roll_gain = config['balance_ankle_roll_gain']
        walking_param.balance_ankle_pitch_gain = config['balance_ankle_pitch_gain']
        walking_param.y_swap_amplitude = config['y_swap_amplitude']
        walking_param.z_swap_amplitude = config['z_swap_amplitude']
        walking_param.arm_swing_gain = config['arm_swing_gain']
        walking_param.pelvis_offset = config['pelvis_offset']
        walking_param.hip_pitch_offset = config['hip_pitch_offset']
        walking_param.p_gain = config['p_gain']
        walking_param.i_gain = config['i_gain']
        walking_param.d_gain = config['d_gain']

        return walking_param

    def __update_walking_gait(self, gait, x_move=None, y_move=None, angle_rot=None):
        extra_params = list(WalkingModule.GAIT_WALKING_PARAM_KEYS - set(self.gaits[gait].keys()))

        walking_gait = dict(self.default_walking_gait)

        # Override default parameters if specified in gait
        for param in extra_params:
            walking_gait[param] = self.gaits[gait][param]

        # Check to make sure all default and overridden gait parameters are valid
        for param, value in walking_gait.iteritems():
            if value == sys.maxint:
                rospy.logerr('Missing walking module default parameter: {0}'.format(param))
            elif param in WalkingModule.DEGREE_WALKING_PARAM_KEYS:
                # Convert all values that are in degrees to radians
                walking_gait[param] = math.radians(value)

        walking_param = self.__set_walking_params(walking_param=WalkingParam(), config=walking_gait)
        walking_param.init_x_offset = self.gaits[gait]['init_x_offset']
        walking_param.init_y_offset = self.gaits[gait]['init_y_offset']
        walking_param.init_z_offset = self.gaits[gait]['init_z_offset']
        walking_param.arm_swing_gain = self.gaits[gait]['arm_swing_gain']
        walking_param.z_move_amplitude = self.gaits[gait]['z_move_amplitude']
        walking_param.period_time = self.gaits[gait]['period_time']
        walking_param.hip_pitch_offset = math.radians(self.gaits[gait]['hip_pitch_offset'])
        
        # if x_move != None and y_move != None and angle_rot != None:
        #     walking_param.y_move_amplitude = y_move
        #     walking_param.angle_move_amplitude = math.radians(angle_rot)
        # else:
        #     walking_param.x_move_amplitude = self.gaits[gait]['x_move_amplitude']
        #     walking_param.y_move_amplitude = self.gaits[gait]['y_move_amplitude']
        #     walking_param.angle_move_amplitude = math.radians(self.gaits[gait]['angle_move_amplitude'])

        if x_move == None:
            walking_param.x_move_amplitude = self.gaits[gait]['x_move_amplitude']
        else: 
            walking_param.x_move_amplitude = x_move

        if y_move == None:
            walking_param.y_move_amplitude = self.gaits[gait]['y_move_amplitude']
        else: 
            walking_param.y_move_amplitude = y_move

        if angle_rot == None:
            walking_param.angle_move_amplitude = math.radians(self.gaits[gait]['angle_move_amplitude'])
        else: 
            walking_param.angle_move_amplitude = math.radians(angle_rot)



        self.__set_walking_param_pub.publish(walking_param)

    def __send_walking_command(self, command):
        message = String()
        message.data = command

        self.__set_walking_command_pub.publish(message)

    def set_initial_walk(self, gait):
        self.is_initial_walk = True
        self.initial_walk = gait

    def walk(self, gait='', stop=False, x_move=None, y_move=None, angle_rot=None):
        if not WalkingModule.walk_enabled:
            rospy.logwarn('Walking module is not enabled')
            return

        if stop:
            rospy.loginfo('Stopping walk.')
            self.__send_walking_command(command='stop')
            WalkingModule.walk_enabled = False
            return

        if gait in self.gaits.keys():
            rospy.loginfo('Switching walking gait: {0}'.format(gait))

            if self.is_initial_walk:
                if self.robot_size == 'kid':
                    self.__update_walking_gait(gait=self.initial_walk)
                    self.__send_walking_command(command='start')
                    self.is_initial_walk = False
                    sleep(1)
                else:
                    self.__send_walking_command(command='start')
                    self.is_initial_walk = False

            self.__update_walking_gait(gait=gait, x_move=x_move, y_move=y_move, angle_rot=angle_rot)
        else:
            rospy.logwarn('Walking gait ({0}) not found in walking_configurations'.format(gait))

    def march(self):
        self.walk(gait='march')

    def forward(self):
        self.walk(gait='forward')

    def backward(self):
        self.walk(gait='backward')
    
    def turn_left(self):
        self.walk(gait='turn_left')

    def turn_right(self):
        self.walk(gait='turn_right')
    
    def follow(self):
        self.walk(gait='follow')

    def stop(self):
        self.walk(stop=True)


