#!/usr/bin/env python

from enum import Enum
from time import sleep

import rosnode
import rospy
import math
from std_msgs.msg import String
from sensor_msgs.msg import JointState

from control_modules.control_module import ControlModule


class ObjectStatus(Enum):
    NOTFOUND = 0
    STRAIGHT = 1
    ONRIGHT = 2
    ONLEFT = 3


class Follower(object):
    MAX_FB_STEP = 40.0 * 0.001
    MIN_FB_STEP = 5.0 * 0.001
    MAX_RL_TURN = 15.0 * math.pi / 180.
    MIN_RL_TURN = 5.0 * math.pi / 180.
    UNIT_RL_TURN = 0.5 * math.pi / 180.

    def __init__(self, event, robot_size, control_module, advanced_control=False):
        self.tracking = False
        self.in_place_fb_step = -3.0 * 0.001
        self.hip_pitch_offset = 7.0
        self.curr_period_time = 0.6
        self.current_pan = -10
        self.current_tilt = -10
        self.count_to_kick = 0
        self.count_not_found = 0
        self.accum_position = 0
        self.accum_period = 0
        self.approach_position = ObjectStatus.NOTFOUND
        self.NOT_FOUND_THRESHOLD = 50
        self.prev_time = rospy.Time.now()
        self.curr_x_move = 0.005
        self.curr_r_angle = 0
        self.event = event
        self.robot_size = robot_size

        self.control_module = control_module
        self.with_advanced = advanced_control

        self.__current_joint_states = rospy.Subscriber('/robotis/goal_joint_states', JointState, self.get_join_states_callback)

    
    def start_following(self):
        self.tracking = True
        rospy.loginfo('Start Following')

        self.set_walking_command('start')

        self.control_module.walking().follow()


    def stop_following(self):
        self.tracking = False
        self.approach_position = ObjectStatus.NOTFOUND
        self.count_to_kick = 0
        self.accum_position = 0
        rospy.loginfo('Stop Following')

        self.set_walking_command('stop')


    def get_join_states_callback(self, msg):
        get_count = 0
        pan, tilt = 0.0, 0.0

        for joint in range(len(msg.name)):
            if msg.name[joint] == 'head_pan':
                pan = msg.position[joint]
                get_count += 1
            elif msg.name[joint] == 'head_tilt':
                tilt = msg.position[joint]
                get_count += 1

            if get_count == 2:
                break
        
        self.current_tilt = tilt
        self.current_pan = pan


    def set_walking_command(self, command):
        if command == 'start':
            self.set_walking_param(self.in_place_fb_step, 0, 0)
        else:
            self.control_module.walking().stop()

    def wait_following(self):
        self.count_not_found += 1

        if self.count_not_found > self.NOT_FOUND_THRESHOLD * 0.5:
            self.set_walking_param(0.0,0.0,0.0)


    def process_following(self, x_angle, y_angle, obj_size, gait = "follow", auto = True):
        curr_time = rospy.Time.now()
        duration = curr_time - self.prev_time
        self.prev_time = curr_time
        delta_time = duration.nsecs * 0.000000001 + duration.secs

        self.count_not_found = 0
        ball_position_sum = 0

        approach_position = ObjectStatus.NOTFOUND
        if self.robot_size == 'kid':
            distance_to_obj = 0.46 * math.tan(math.pi * 0.5 + self.current_tilt - self.hip_pitch_offset - obj_size)
        else:
            # head measurement is 0.88 for adult
            distance_to_obj = 0.88 * math.tan(math.pi * 0.5 + self.current_tilt - self.hip_pitch_offset - obj_size)

        obj_y_angle = (self.current_tilt + y_angle) * 180 / math.pi
        obj_x_angle = (self.current_pan + x_angle) * 180 / math.pi

        if distance_to_obj < 0:
            distance_to_obj *= -1.
            
        distance_to_kick = 0.22
            
        fb_move, rl_angle = 0, 0
        distance_to_walk = distance_to_obj - distance_to_kick


        
        fb_move, rl_angle = self.calc_footsteps(gait, distance_to_walk, self.current_pan, delta_time, fb_move, rl_angle)

        # if self.event == 'sprint':
        #     if gait == 'follow':
        #         if self.robot_size == 'kid':
        #             fb_move = 0.027
        #         else:
        #             fb_move = 0.027
        #     else:
        #         if self.robot_size == 'kid':
        #             fb_move = -0.025
        #         else:
        #             fb_move = -0.025

        
        if not auto:
            fb_move = None

        self.set_walking_param(fb_move, 0, rl_angle, gait)

        return rl_angle

    def set_walking_param(self, x_move, y_move, rot_angle, gait = 'follow'):
        self.curr_x_move = x_move
        self.curr_r_angle = rot_angle
        angle_rot = math.degrees(rot_angle)
        if self.with_advanced:
            if angle_rot > 0:
                self.control_module.walking().turn_left()
            elif angle_rot < 0:
                self.control_module.walking().turn_right()
            else:
                self.control_module.walking().forward()
        else:
            self.control_module.walking().walk(gait= gait, stop=False, x_move=x_move, y_move=y_move, angle_rot=angle_rot)
        #elif self.event == 'sprint':
            #self.control_module.walking().walk(gait= gait, stop=False, x_move=x_move, y_move=y_move,angle_rot=angle_rot)

    def calc_footsteps(self, gait, target_distance, target_angle, delta_time, fb_move, rl_angle):
        next_move = self.curr_x_move

        if target_distance < 0:
            target_distance = 0.

        fb_goal = min(target_distance*0.1, Follower.MAX_FB_STEP)

        fb_goal = min(next_move, fb_goal)
        fb_move = max(fb_goal, Follower.MIN_FB_STEP)

        rl_goal = 0.

        if math.fabs(target_angle) * 180 / math.pi > 5.0:

            if self.robot_size == 'kid':
                if gait == 'follow':
                    rl_offset = math.fabs(target_angle) * 0.2
                else:
                    rl_offset = math.fabs(target_angle) * -0.1
            else:
                if gait == 'follow':
                    rl_offset = math.fabs(target_angle) * 0.1
                else:
                    rl_offset = math.fabs(target_angle) * -0.1

            rl_goal = min(rl_offset, Follower.MAX_RL_TURN)
            rl_goal = max(rl_goal, Follower.MIN_RL_TURN)
            rl_angle = min(math.fabs(self.curr_r_angle) + Follower.UNIT_RL_TURN, rl_goal)

            if target_angle < 0:
                rl_angle *= -1

        return fb_move, rl_angle