#!/usr/bin/env python

import json
import os
import math
from time import sleep
from enum import Enum

import roslib.packages as rospkg
import rospy
from std_msgs.msg import Int32, String

from control_modules.control_module import ControlModule
from vision.msg import PositionArea
from events.msg import Track
from sensor_msgs.msg import JointState


class TrackingStatus(Enum):
    NOTFOUND = -1
    WAITING = 0
    FOUND = 1


class Tracker(object):
    def __init__(self, event, robot_size, control_module):
        self.use_head_scan = True
        self.on_tracking = False
        self.curr_pan = 0
        self.curr_tilt = 0
        self.curr_bottom = 0
        self.x_error_sum = 0
        self.y_error_sum = 0
        self.count_not_found = 0
        self.tracking_status = TrackingStatus.NOTFOUND
        # From vision:
        self.x = -1
        self.y = -1
        self.area = -1
        self.WAITING_THRESHOLD = 5
        self.NOT_FOUND_THRESHOLD = 50
        self.FOV_WIDTH = 35.2 * math.pi / 180
        self.FOV_HEIGHT = 21.6 * math.pi / 180
        self.prev_time = rospy.Time.now()

        # Initializing modules
        self.control_module = control_module

        self.__set_head_scan_pub = rospy.Publisher(
            '/robotis/head_control/scan_command', String, queue_size=0)

        self.__set_head_joint_pub = rospy.Publisher(
            '/robotis/head_control/set_joint_states_offset', JointState, queue_size=0)

        self.track_sub = rospy.Subscriber('/track', Track, self.track_callback)

    
    def track_callback(self, msg):
        # This one uses the suscriber's values from vision
        if msg.size >= 0:
            # The following equation is the one robotis use for ball_detector
            # in order to make the tracker work, we need to convert the position
            # to the scale of -1 to 1
            self.x = msg.x / 427*2 - 1
            self.y = msg.y / 320*2 - 1
            sleep(0.001)
            self.area = msg.size

    def provide_marker(self, x, y, area):
        # This one is being passed the values directly. Seems faster
        # The following equation is the one robotis use for ball_detector
        # in order to make the tracker work, we need to convert the position
        # to the scale of -1 to 1
        self.x = x / 427*2 - 1
        self.y = y / 320*2 - 1
        sleep(0.001)
        if not (x<=-1 and y<=-1) :
            self.area = area
        else:
            self.area = -1

    
    def tracker_command_callback(self, msg):
        if msg.data == 'start':
            self.start_tracking()
        elif msg.data == 'stop':
            self.stop_tracking()
        elif msg.data == 'toggle_start':
            if self.on_tracking:
               self.start_tracking()
            else:
                self.stop_tracking() 


    def process_tracking(self):
        status = TrackingStatus.FOUND

        if not self.on_tracking:
            self.area = 0
            self.count_not_found = 0
            return TrackingStatus.NOTFOUND

        # Check object position: ----------------------

        if self.area <= 0:
            self.count_not_found += 1
            if self.count_not_found < self.WAITING_THRESHOLD:
                if self.tracking_status == TrackingStatus.FOUND or self.tracking_status == TrackingStatus.WAITING:
                    status = TrackingStatus.WAITING
                else:
                    status = TrackingStatus.NOTFOUND
            elif self.count_not_found > self.NOT_FOUND_THRESHOLD:
                self.scan_object()
                self.count_not_found = 0
                status = TrackingStatus.NOTFOUND
            else:
                status = TrackingStatus.NOTFOUND
        else:
            self.count_not_found = 0

        # Convert position to angle: -------------------------

        # if object is found
        # convert object position to desired angle(rad) of head
        # obj_position : top-left is (-1, -1), bottom-right is (+1, +1)
        # offset_rad : top-left(+, +), bottom-right(-, -)

        x_error, y_error, size = 0.0, 0.0, 0.0

        if status == TrackingStatus.NOTFOUND:
            self.tracking_status = status
            self.curr_pan = 0
            self.curr_tilt = 0
            self.x_error_sum = 0
            self.y_error_sum = 0
            return status
        elif status == TrackingStatus.WAITING:
            self.tracking_status = status
            return status
        elif status == TrackingStatus.FOUND:
            x_error = -math.atan(self.x * math.tan(self.FOV_WIDTH))
            y_error = -math.atan(self.y * math.tan(self.FOV_HEIGHT))
            size = self.area

        curr_time = rospy.Time.now()
        duration = curr_time - self.prev_time
        self.prev_time = curr_time

        delta_time = duration.nsecs * 0.000000001 + duration.secs

        p_gain = 0.4
        i_gain = 0.0
        d_gain = 0.0

        x_error_diff = (x_error - self.curr_pan) 
        y_error_diff = (y_error - self.curr_tilt) 

        x_error_target = x_error * p_gain + x_error_diff * d_gain
        y_error_target = y_error * p_gain + y_error_diff * d_gain

        # Move head: -------------------

        self.publish_head_joint(pan=x_error_target, tilt=y_error_target)

        # Args for following object: ----------------

        self.curr_pan = x_error
        self.curr_tilt = y_error
        self.curr_bottom = size

        self.area = 0
        self.tracking_status = status

        return status


    def publish_head_joint(self, pan, tilt):
        min_angle = 1 * math.pi / 180.
        if not (math.fabs(pan) < min_angle and math.fabs(tilt) < min_angle):
            self.control_module.head()
            msg = JointState()
            msg.name.append('head_pan')
            msg.name.append('head_tilt')

            msg.position.append(pan)
            msg.position.append(tilt)

            self.__set_head_joint_pub.publish(msg)
            sleep(0.01)        

    def start_tracking(self):
        self.on_tracking = True
        rospy.loginfo('Start Tracker')


    def stop_tracking(self):
        self.on_tracking = False
        rospy.loginfo('Stop Tracker')

        self.curr_pan = 0
        self.curr_tilt = 0
        self.x_error_sum = 0
        self.y_error_sum = 0

    
    def get_pan(self):
        return self.curr_pan

    
    def get_tilt(self):
        return self.curr_tilt

    
    def get_size(self):
        return self.curr_bottom


    def set_using_head_scan(self, use_scan):
        self.use_head_scan = use_scan
    

    def scan_object(self):
        if self.use_head_scan:
            self.control_module.head()
            self.__set_head_scan_pub.publish('scan') # call scan from head_control_module.cpp
    