#!/usr/bin/env python

from enum import Enum
from time import sleep

import rosnode
import rospy
import roslib.packages as rospkg
import os
import json
import math
from std_msgs.msg import String

from control_modules.control_module import AdvancedControlModule
from soccer_obs import Obstacles
from vision.msg import PositionArea, DetectedLines
from tools.follower import Follower
from tools.tracker import Tracker
from events.msg import Track
from slam.srv import ToggleLocalization
from slam.msg import RobotPosition
from robot import Robot, RobotMovement


class RobotMode(Enum):
    READY = 0
    RUNNING = 1

class RobotStart(Enum):
    LEFT = 0
    CENTER = 1
    RIGHT = 2


class Movements(Enum):
    MARCH = 0
    FORWARD = 1
    BACKWARD = 2
    TURN_LEFT = 3
    TURN_RIGHT = 4
    ACTION = 5


class StandState(Enum):
    STAND = 0
    FALLEN_FORWARD = 1
    FALLEN_BACKWARD = 2


def button_callack(msg):
    global robot_mode, mode_change, stopped, robot_start_pos

    
    if msg.data == 'mode':
        robot_start_pos = RobotStart.LEFT
        if with_slam:
            toggle_localization_srv(place_left=True, start=False, pause=True, obstacle=False, turn_left=False, turn_right=False, place_center=False, place_right=False)
        rospy.loginfo('Positioning at LEFT of field')
    elif msg.data == 'start':
        robot_start_pos = RobotStart.CENTER
        if with_slam:
            toggle_localization_srv(place_center=True, start=False, pause=True, obstacle=False, turn_left=False, turn_right=False, place_left=False, place_right=False)
        rospy.loginfo('Positioning at CENTER of field')
    elif msg.data == 'user':
        robot_start_pos = RobotStart.RIGHT 
        if with_slam:
            toggle_localization_srv(place_right=True, start=False, pause=True, obstacle=False, turn_left=False, turn_right=False, place_left=False, place_center=False)
        rospy.loginfo('Positioning at RIGHT of field')
    elif msg.data == 'start_long':
        # Start button pressed
        # Send robot to RUNNING mode
        if robot_mode == RobotMode.READY:
            robot_mode = RobotMode.RUNNING
            mode_change = True
        elif robot_mode == RobotMode.RUNNING:
            rospy.logerr('Robot already in running mode.')
    elif msg.data == 'mode_long':
            # Mode button pressed
            # Send robot to READY mode
            robot_mode = RobotMode.READY
            mode_change = True
            stopped = True



def wait_for_node(node_name):
    # Wait for node to start
    while node_name not in rosnode.get_node_names():
        rospy.logwarn('Node is not running: {0}'.format(node_name))
        sleep(1)

    rospy.loginfo('Node is running: {0}'.format(node_name))


def memorize_lines(msg):
    global lines, look_at_lines

    if look_at_lines:
        if 'T' in msg.boundary_line and 'L' in msg.goal_line:
            lines += ['TL']
        elif 'L' in msg.goal_line:
            lines += ['gL']
        elif 'center' in msg.center_line:
            lines += ['center']
        else:
            lines += ['']


def split_list(alist, wanted_parts=1):
    length = len(alist)
    return [alist[i*length // wanted_parts: (i+1)*length // wanted_parts] for i in range(wanted_parts)]



def play_soccer():
    global start_following, following_ball, object_found, obj_tracker, track_pub, obstacles, with_slam, angle_robot, robot_size

    movement = None
    previous_movement = None

    count = 0
    angle_robot = 0
    stage = 0
    count_lost = 0
    last_saved_angle = 0

    while not rospy.is_shutdown() and not mode_change:
        tracking_status = obj_tracker.process_tracking()

        if with_slam:            
            print(angle_robot)
            if last_saved_angle > 0:
                if angle_robot >= 0 and angle_robot < 45:
                    toggle_localization_srv(pause=True, turn=True, turn_to_angle=6)
                elif angle_robot >= 45 and angle_robot < 90:
                    toggle_localization_srv(pause=True, turn=True, turn_to_angle=5)
                elif angle_robot >= 90 and angle_robot < 135:
                    toggle_localization_srv(pause=True, turn=True, turn_to_angle=4)
                elif angle_robot >= 135 and angle_robot < 180:
                    toggle_localization_srv(pause=True, turn=True, turn_to_angle=3)
                elif angle_robot >= 180 and angle_robot < 225:
                    toggle_localization_srv(pause=True, turn=True, turn_to_angle=2)
                elif angle_robot >= 225 and angle_robot < 270:
                    toggle_localization_srv(pause=True, turn=True, turn_to_angle=1)
                elif angle_robot >= 270 and angle_robot < 315:
                    toggle_localization_srv(pause=True, turn=True, turn_to_angle=0)
                elif angle_robot >= 315:
                    toggle_localization_srv(pause=True, turn=True, turn_to_angle=7)
            elif last_saved_angle < 0:
                if angle_robot <= 0 and angle_robot > -45:
                    toggle_localization_srv(pause=True, turn=True, turn_to_angle=6)
                elif angle_robot <= -45 and angle_robot > -90:
                    toggle_localization_srv(pause=True, turn=True, turn_to_angle=7)
                elif angle_robot <= -90 and angle_robot > -135:
                    toggle_localization_srv(pause=True, turn=True, turn_to_angle=0)
                elif angle_robot <= -135 and angle_robot > -180:
                    toggle_localization_srv(pause=True, turn=True, turn_to_angle=1)
                elif angle_robot <= -180 and angle_robot > -225:
                    toggle_localization_srv(pause=True, turn=True, turn_to_angle=2)
                elif angle_robot <= -225 and angle_robot > -270:
                    toggle_localization_srv(pause=True, turn=True, turn_to_angle=3)
                elif angle_robot <= -270 and angle_robot > -315:
                    toggle_localization_srv(pause=True, turn=True, turn_to_angle=4)
                elif angle_robot <= -315:
                    toggle_localization_srv(pause=True, turn=True, turn_to_angle=5)
            else:
                if angle_robot > 45:
                    toggle_localization_srv(pause=True, turn=True, turn_to_angle=5)
                elif angle_robot < -45:
                    toggle_localization_srv(pause=True, turn=True, turn_to_angle=7)

            last_saved_angle = angle_robot
            toggle_localization_srv(pause=False)


            movement = robot.next_movement(avoid_obstacle=obstacles.blue.is_center, num_map_rows=NUM_MAP_ROWS, num_map_columns=NUM_MAP_COLUMNS)

            
            if movement == RobotMovement.OBSTACLE:
                # Avoid obstacle right in front of robot with localization turned off
                rospy.loginfo('[Navigation] avoid obstacle')

                toggle_localization_srv(pause=True, obstacle=True, start=False, turn_left=False, turn_right=False, place_left=False, place_right=False, place_center=False)
                control_module.walking().turn_right()
                sleep(robot.turn_right_rate / 4.0)
                control_module.walking().forward()
                sleep(robot.forward_rate * 8)
                control_module.walking().turn_left()
                sleep(robot.turn_left_rate / 4.0)
                control_module.walking().stop()
                toggle_localization_srv(pause=False)

            elif movement == RobotMovement.TURN_LEFT:
                obj_tracker.stop_tracking()
                obj_follower.stop_following()
                # Turn left
                rospy.loginfo('[Navigation] turn left')

                toggle_localization_srv(pause=True, turn_left90=True, start=False, obstacle=False, turn_right=False, place_left=False, place_right=False, place_center=False)
                control_module.walking().turn_left()
                sleep(robot.turn_left_rate)
                control_module.walking().stop()
                toggle_localization_srv(pause=False)
                start_following = True

            elif movement == RobotMovement.TURN_RIGHT:
                obj_tracker.stop_tracking()
                obj_follower.stop_following()
                # Turn right
                rospy.loginfo('[Navigation] turn right')

                toggle_localization_srv(pause=True, turn_right90=True, start=False, turn_left=False, obstacle=False, place_left=False, place_right=False, place_center=False)
                control_module.walking().turn_right()
                sleep(robot.turn_right_rate)
                control_module.walking().stop()
                toggle_localization_srv(pause=False)
                start_following = True

            previous_movement = movement


        obj_tracker.provide_marker(obstacles.circle.x, obstacles.circle.y, obstacles.circle.area)
        

        if start_following:
            obj_tracker.start_tracking()
            obj_follower.start_following()
            start_following = False
            sleep(1)


        if stage == 0:

            if obstacles.circle.area > min_ball_size:
                if obstacles.circle.area < max_ball_size:
                    obj_follower.process_following(obj_tracker.get_pan(), obj_tracker.get_tilt(), 0)
                    count = 0
                    angle_robot = control_module.plan_orientation['angle']
                else:
                    count += 1
                    if count > 3:
                        obj_tracker.stop_tracking()
                        obj_follower.stop_following()
                        stage = 1
                        rospy.logwarn('[United Soccer] Stage {0}'.format(stage))
                        count = 0
            else:
                count = 0
                obj_follower.wait_following()
        elif stage == 1:
            control_module.head().move_head(pan_position=0, tilt_position=-0.9)
            sleep(1)
            angle_robot = control_module.plan_orientation['angle']
            rospy.logwarn('turning {}'.format(angle_robot))
            stage = 2
            rospy.logwarn('[United Soccer] Stage {0}'.format(stage))
        elif stage == 2:
            if angle_robot < 0:
                # if obstacles.circle.x > 150:
                #     control_module.walking('shift_right').shift_right()
                #     sleep(1)
                # elif obstacles.circle.x > 0 and obstacles.circle.x < 150:
                control_module.walking().turning_helper(mode='turn_left_inplace',angle=float(angle_robot))
                stage = 3
                rospy.logwarn('[United Soccer] Stage {0}'.format(stage))
            else:
                # if obstacles.circle.x < 277:
                #     control_module.walking('shift_left').shift_left()
                #     sleep(1)
                # else:
                control_module.walking().turning_helper(mode='turn_right_inplace', angle=float(angle_robot))
                stage = 3
                rospy.logwarn('[United Soccer] Stage {0}'.format(stage))
            

        elif stage == 3:
            control_module.head().move_head(pan_position=0, tilt_position=-1.0)
            if obstacles.circle.y > 0 and obstacles.circle.y < 280:
                control_module.walking().forward()
                count_lost = 0
                sleep(0.5)
            elif obstacles.circle.y < 0:
                count_lost += 1
                if count_lost > 10:
                    control_module.walking().backward()
                    sleep(0.5)
                    start_following = True
                    stage = 0
                    count_lost = 0
            elif obstacles.circle.x < 150:
                control_module.walking().walk(gait='shift_left')
                sleep(1)
                count_lost = 0
            elif obstacles.circle.x > 277:
                control_module.walking().walk(gait='shift_right')
                sleep(1)
                count_lost = 0
            
            else:
                handle_kick(obstacles.circle.x)
                stage = 0
                count_lost = 0
                rospy.logwarn('[United Soccer] Stage {0}'.format(stage))
                start_following = True


        stand_state = control_module.check_fallen()
        if stand_state != 'STAND':
            obj_follower.stop_following()
            handle_fallen(stand_state)
            start_following = True



def handle_kick(obj_position):
    control_module.head().move_head(pan_position=0, tilt_position=-1.0)
    sleep(0.5)
    print('ball x', obj_position)
    if obj_position > 213: # 213 is half of resized vision resolution
        control_module.action().play_action('right_kick')
    else:
        control_module.action().play_action('left_kick')
    control_module.action().play_action('walk_ready')
    sleep(1)
    control_module.head().move_head(pan_position=0, tilt_position=-1.0)



def handle_fallen(status):
    if status == 'FALLEN_FORWARD':
        control_module.action().play_action('get_up_front')
    elif status == 'FALLEN_BEHIND':
        control_module.action().play_action('get_up_back')
    control_module.action().play_action('walk_ready')
    control_module.head().move_head(pan_position=0, tilt_position=-1.0)
    
    sleep(2)


def set_variables():
    global min_ball_size, max_ball_size
    config_variables_file = os.path.join(rospkg.get_pkg_dir('events'), 'config', event, 'variables_{0}.json'.format(robot_size))

    vars = None
    with open(config_variables_file, 'r') as fp:
            vars = json.load(fp)
    
    min_ball_size = int(vars['min_ball_size'])
    max_ball_size = int(vars['max_ball_size'])


if __name__ == '__main__':

    # Initalize global variables
    robot_mode = RobotMode.READY
    mode_change = False
    robot = Robot(forward_rate=1.0, left_rate = 10.0, right_rate = 9.0)
    obstacles = Obstacles()
    SPIN_RATE = 0.5
    NUM_MAP_ROWS, NUM_MAP_COLUMNS = 19, 28
    event = 'united_soccer'
    start_following = True
    look_at_lines = False
    stopped = False
    robot_start_pos = RobotStart.CENTER

    lines = []

    with_slam = rospy.get_param('united_soccer_node/with_slam')
    robot_size = rospy.get_param('united_soccer_node/size')

    set_variables() # Important !

    # Wait for other nodes to launch
    wait_for_node('/op3_manager')
    wait_for_node('/vision_node')
    if with_slam:
        wait_for_node('/slam_node')

    # Initialze ROS node
    rospy.init_node('united_soccer_node')
    rate = rospy.Rate(SPIN_RATE)

    # Initialize OP3 ROS generic subscriber
    button_sub = rospy.Subscriber('/robotis/open_cr/button', String, button_callack)

    # Initialze vision subscribers
    line_vision_sub = rospy.Subscriber('/vision_node/lines', DetectedLines, memorize_lines)
    blue_sub = rospy.Subscriber('/vision_node/blue', PositionArea,
                                lambda msg: obstacles.blue.update(msg.x, msg.y, msg.position, msg.area))
    red_sub = rospy.Subscriber('/vision_node/red', PositionArea,
                               lambda msg: obstacles.red.update(msg.x, msg.y, msg.position, msg.area))
    yellow_sub = rospy.Subscriber('/vision_node/yellow', PositionArea,
                                        lambda msg: obstacles.yellow.update(msg.x, msg.y, msg.position, msg.area))
    mixed_sub = rospy.Subscriber('/vision_node/mixed', PositionArea,
                                        lambda msg: obstacles.mixed.update(msg.x, msg.y, msg.position, msg.area))
    circle_sub = rospy.Subscriber('/vision_node/circle', PositionArea,
                                        lambda msg: obstacles.circle.update(msg.x, msg.y, msg.position, msg.area))
        

    
    track_pub = rospy.Publisher('/track', Track, queue_size=0)

    # SLAM:
    if with_slam:
        toggle_localization_srv = rospy.ServiceProxy('/slam/toggle_localization', ToggleLocalization)
        robot_position_sub = rospy.Subscriber('/slam/robot_position', RobotPosition, robot.update_pose)

    

    # Initializing modules
    control_module = AdvancedControlModule(event, robot_size)

    obj_tracker = Tracker(event, robot_size, control_module)
    obj_follower = Follower(event, robot_size, control_module, True)

    # Enabling walking module
    sleep(10)

    rospy.loginfo('Positioning at CENTER of field by default')

    control_module.action().play_action('walk_ready')
    sleep(0.5)
    control_module.head().move_head(pan_position=0, tilt_position=-1.0)

    if robot_size == 'adult':
        robot_mode = RobotMode.RUNNING
        mode_change = True
    elif robot_size == 'kid':
        rospy.loginfo('Press start button to begin.')

    while not rospy.is_shutdown():
        if mode_change:
            mode_change = False

            if robot_mode == RobotMode.READY:
                rospy.loginfo('Resetting OP3. Press start button to begin.')

                control_module.walking().stop()
                control_module.head().move_head(pan_position=0, tilt_position=-0.85)
            
            elif robot_mode == RobotMode.RUNNING:
                rospy.loginfo('Starting united soccer event.')

                if with_slam:
                    # # look left
                    # control_module.head().move_head(pan_position=0.75, tilt_position=-0.8)
                    # sleep(1.5)
                    # look_at_lines = True
                    # # look right
                    # control_module.head().move_head(pan_position=-0.75, tilt_position=-0.8)
                    # sleep(2.5)
                    # look_at_lines = False

                    # left_lines, center_lines, right_lines = split_list(lines, wanted_parts=3)
                    # lines = [max(left_lines, key=left_lines.count),
                    #         max(center_lines, key=center_lines.count),
                    #         max(right_lines, key=right_lines.count)]

                    # if lines[0] == 'TL':
                    #     toggle_localization_srv(place_left=True, start=False, pause=True, obstacle=False, turn_left=False, turn_right=False, place_center=False, place_right=False)
                    # elif lines[2] == 'TL':
                    #     toggle_localization_srv(place_right=True, start=False, pause=True, obstacle=False, turn_left=False, turn_right=False, place_left=False, place_center=False)
                    # else:
                    #     toggle_localization_srv(place_center=True, start=False, pause=True, obstacle=False, turn_left=False, turn_right=False, place_left=False, place_right=False)
                    toggle_localization_srv(start=True)

                    control_module.head().move_head(pan_position=0, tilt_position=-0.85)


                play_soccer()

        rate.sleep()
