
import cv2 as cv
from frame import Frame
import copy
import opencv_gui
import sys
import json
# author: Borui Li


class Video:
    def __init__(self, config_path, camera_index=0):
        self.stream = cv.VideoCapture(camera_index)
        self.input_frame = None
        self.output_frame = None
        self.raw_frame = None
        self.config = None
        self.is_frozen = False
        self.frozen_frame = None

        with open(config_path, 'r') as fp:
            self.config = json.load(fp)

        self.opencv_gui = opencv_gui.gui(self.config, config_path)
        self.markers = []
        self.current_marker_name = self.config['markers'].keys()[0]
        self.mouse_selected = (-1, -1)
        self.is_mouse_down = False

    def froze_frame(self):
        self.is_frozen = not self.is_frozen
        if self.is_frozen:
            self.frozen_frame = copy.deepcopy(self.raw_frame)

    def next_frame(self):

        if not self.is_frozen:
            _, self.raw_frame = self.stream.read()
        else:
            self.raw_frame = copy.deepcopy(self.frozen_frame)

        if self.raw_frame is not None and self.raw_frame.data:
            self.input_frame = Frame(self.raw_frame, width=self.config['frame_width'],
                                     height=self.config['frame_height'])
            self.output_frame = self.input_frame.clone()
            return True
        else:
            return False

    def initial(self):
        self.initial_track_bar()
        self.mouse_call_back()

    def call_back(self, name, new_value):
        self.config['markers'][self.current_marker_name][name] = new_value
        self.opencv_gui.save_json(self.config)

    def mouse_event(self, event, x, y):
        if event == cv.EVENT_LBUTTONDOWN and not self.is_mouse_down:
            self.mouse_selected = (x, y)
            self.is_mouse_down = True
        elif event == cv.EVENT_MOUSEMOVE and self.is_mouse_down:
            self.draw_rect(self.input_frame, self.mouse_selected,
                           (x, y), (0, 255, 0))
            # print (self.mouse_selected,(x,y))
            self.show_output()
        elif event == cv.EVENT_LBUTTONUP:
            self.is_mouse_down = False
            if self.mouse_selected != (x, y):
                self.config['markers'][self.current_marker_name]['selected_region']['x'] = self.mouse_selected[0]
                self.config['markers'][self.current_marker_name]['selected_region']['y'] = self.mouse_selected[1]
                self.config['markers'][self.current_marker_name]['selected_region']['width'] = x - \
                    self.mouse_selected[0]
                self.config['markers'][self.current_marker_name]['selected_region']['height'] = y - \
                    self.mouse_selected[1]

            print('set marker:')
            print('left_top:{0} right_bottom:{1}, width:{3}, height: {4}'.format(
                self.config['markers'][self.current_marker_name]['selected_region']['x'],
                self.config['markers'][self.current_marker_name]['selected_region']['y'],
                x, y,
                self.config['markers'][self.current_marker_name]['selected_region']['width'],
                self.config['markers'][self.current_marker_name]['selected_region']['height'])
            )
            self.opencv_gui.save_json(self.config)

    def mouse_call_back(self):
        self.opencv_gui.set_mouse_call_back(self.opencv_gui.output_window_name, lambda event, x, y, flags, params:
                                            self.mouse_event(event=event, x=x, y=y))

    def initial_track_bar(self):
        self.opencv_gui.create_track_bar(self.opencv_gui.output_window_name, 'bright', 0, 255,
                                         lambda x: self.call_back('bright', x))
        self.opencv_gui.create_track_bar(self.opencv_gui.output_window_name, 'erode', 0, 10,
                                         lambda x: self.call_back('erode', x))
        self.opencv_gui.create_track_bar(self.opencv_gui.output_window_name, 'dilate', 0, 10,
                                         lambda x: self.call_back('dilate', x))
        self.opencv_gui.create_track_bar(self.opencv_gui.output_window_name, 'pattle_thresh', 0, 20,
                                         lambda x: self.call_back('pattle_thresh', x))
        self.opencv_gui.set_tackbar_pos(self.opencv_gui.output_window_name,
                                        'bright', self.config['markers'][self.current_marker_name]['bright'])
        self.opencv_gui.set_tackbar_pos(self.opencv_gui.output_window_name,
                                        'erode', self.config['markers'][self.current_marker_name]['erode'])
        self.opencv_gui.set_tackbar_pos(self.opencv_gui.output_window_name,
                                        'dilate', self.config['markers'][self.current_marker_name]['dilate'])
        self.opencv_gui.set_tackbar_pos(self.opencv_gui.output_window_name, 'pattle_thresh',
                                        self.config['markers'][self.current_marker_name]['pattle_thresh'])

    def draw_rect(self, frame, pt1, pt2, color):
        cv.rectangle(frame.frame, pt1, pt2, color, 2)

    def draw_line(self, frame, pt1, pt2, color):
        cv.line(frame.frame, pt1, pt2, color, 2)

    def show_output(self):
        self.opencv_gui.show(
            self.opencv_gui.output_window_name, self.input_frame.frame)

