import cv2 as cv

import opencv_gui as gui
import rospy

from configuration import Configuration
from mat import Mat


class BaseModule(Configuration):

    def __init__(self, configuration_directory, event, robot_size=''):
        super(BaseModule, self).__init__(
            configuration_directory=configuration_directory, event=event, module='base')

        # Initialize camera
        self.cap = cv.VideoCapture(rospy.get_param('vision_node/cam_index'))

        self.original_img = None
        self.output_img = None
        self.debug_objects = []
        self.selected_debug = None

    def add_debug_objects(self, objects):
        self.debug_objects += objects

    def create_trackbars(self):
        self.selected_debug = self.debug_objects[0]

        if len(self.debug_objects) != -1:
            gui.create_trackbar(gui.DEBUG_WINDOW_NAME, trackbar_name='Object', default_value=0,
                                max_value=len(self.debug_objects) - 1,
                                callback=self.switch_debug_object_cb)
        gui.set_mouse_cb('camera', lambda event, x, y, flags, params:
                         self.get_region_colour_space_values_cb(event=event, x=x, y=y))
        gui.create_colour_space_threshold_trackbar(
            default_colour_space_threshold=self.selected_debug.threshold,
            callback=self.set_colour_space_threshold_cb)
        gui.create_opening_closing_ksize_trackbar(
            default_kernel_size=self.selected_debug.opening_closing_ksize,
            callback=self.set_opening_closing_ksize_cb
        )
        gui.create_blur_ksize_trackbar(
            default_kernel_size=self.selected_debug.blur_ksize,
            callback=self.set_blur_ksize_cb
        )

    def read_from_camera(self):
        _, raw_image = self.cap.read()

        if raw_image.data:
            # Create a new Mat object with the raw image
            self.original_img = Mat(
                raw_image, height=self.config['resized_frame_height'])

            # Create an output frame from a clone of the original
            self.output_img = self.original_img.clone()

        return raw_image.data

    def show_debug(self, obj, mat):
        # Show object in debug window if selected
        if self.selected_debug == obj:
            gui.show_debug_window(mat.frame, obj.name)

    def switch_debug_object_cb(self, value):
        # Switch the debug object
        self.selected_debug = self.debug_objects[value]
        gui.set_colour_space_threshold_trackbar_position(
            self.selected_debug.threshold)

    def get_region_colour_space_values_cb(self, event, x, y):
        # Retrieve average values for colour space
        if event == cv.EVENT_LBUTTONDOWN:
            self.colourspace_roi = (x, y)
        elif event == cv.EVENT_LBUTTONUP:
            roi_mean = cv.mean(
                self.original_img.colour_space_frame[self.colourspace_roi[1]:y, self.colourspace_roi[0]:x])
            self.selected_debug.set_colour_space_value(roi_mean[:-1])
            self.selected_debug.module.update()
            self.colourspace_roi = None

    def set_colour_space_threshold_cb(self, threshold):
        # Change colour space threshold for selected object
        self.selected_debug.set_colour_space_threshold(
            threshold=threshold)
        self.selected_debug.module.update()


    def set_opening_closing_ksize_cb(self, kernel_size):
        self.selected_debug.set_opening_closing_ksize(kernel_size)
        self.selected_debug.module.update()

    def set_blur_ksize_cb(self, kernel_size):
        self.selected_debug.set_blur_ksize(kernel_size)
        self.selected_debug.module.update()
