import cv2 as cv
import numpy as np
import os

from base.base_module import BaseModule
from base.configuration import Configuration
import base.opencv_gui as gui

from marker import Marker
from marker_mat import MarkerMat

import roslib.packages as rospkg

class MarkerModule(Configuration):
    def __init__(self, base_module, configuration_directory, event, robot_size):
        super(MarkerModule, self).__init__(
            configuration_directory=configuration_directory, event=event, module='marathon_marker_detector', robot_size=robot_size)

        self.arrow = Marker(configuration=self.config['arrow'], module=self)
        
        self.base = base_module

        self.colourspace_roi = None

        
        # Load template

        folder = 'src/marathon_marker_detector/'

        self.temp_left = cv.imread(os.path.join(rospkg.get_pkg_dir('vision'),'{0}left.png'.format(folder)), cv.IMREAD_GRAYSCALE)
        self.temp_right = cv.imread(os.path.join(rospkg.get_pkg_dir('vision'),'{0}right.png'.format(folder)), cv.IMREAD_GRAYSCALE)
        self.temp_forward = cv.imread(os.path.join(rospkg.get_pkg_dir('vision'),'{0}forward.png'.format(folder)), cv.IMREAD_GRAYSCALE)

        ret, self.temp_left = cv.threshold(self.temp_left, 0, 255, cv.THRESH_BINARY + cv.THRESH_OTSU)
        ret, self.temp_right = cv.threshold(self.temp_right, 0, 255, cv.THRESH_BINARY + cv.THRESH_OTSU)
        ret, self.temp_forward = cv.threshold(self.temp_forward, 0, 255, cv.THRESH_BINARY + cv.THRESH_OTSU)


    @staticmethod
    def is_enabled(base):
        return 'marathon_marker_detector' in base.config['modules']

    def update(self):
        self.config['arrow'] = self.arrow.export_configuration()
        self.save()

    def get_dist(self, p0, p1):
        return ((p1[0] - p0[0]) * (p1[0] - p0[0]) + (p1[1] - p0[1]) * (p1[1] - p0[1])) ** 0.5

    
    def findArrow(self, input_mat, output_mat):

        self.arrow.mat = MarkerMat(frame=input_mat.frame)
        self.arrow.mat.extract_marker(thresh_lb=self.arrow.lower_bound, thresh_ub=self.arrow.upper_bound)
        
        arrow_type = 'None'
        
        global gray, canny, patchCorners, contours, hierarchy, result, sign_roi 

        sign_roi = np.zeros((60,60))

        # flush here??

        gray = cv.cvtColor(input_mat.frame, cv.COLOR_BGR2GRAY)

        canny = cv.Canny(gray, 100, 255, 3)

        # Find all contours from canny edges

        im2, contours, hierarchy = cv.findContours(canny, cv.RETR_TREE, cv.CHAIN_APPROX_SIMPLE)

        # Get the moments
        mu = [None] * len(contours)

        for i in range(len(contours)):
            mu[i] = cv.moments(contours[i], False)
        
        mc = [None] * len(contours)

        # Get the mass centres, use contours to match polygons 
        
        for i in range(len(contours)):
            patch_size = mu[i]['m00']

            check_sign = False # Rectangle of checks

            if patch_size > self.arrow.min_area and patch_size < self.arrow.max_area:
                
                mc[i] = (mu[i]['m10'] / patch_size, mu[i]['m01'] / patch_size)

                approx_arrow = cv.approxPolyDP(contours[i],cv.arcLength(contours[i], True) * 0.05, True)

                # Arrow sign with 4 points
                if len(approx_arrow) == 4:
                    # sort vector using myobject as comparato
                    approx_arrow = self.order_points(approx_arrow)

                    patch_corners = []
                    # arrange the corner position if the drawing pad is not align properly, only if skew to the left
                    # clockwise order from top-left
                    if approx_arrow[0][0] > approx_arrow[1][0]:
                        patch_corners.append(approx_arrow[1])
                        patch_corners.append(approx_arrow[0])
                        patch_corners.append(approx_arrow[2])
                        patch_corners.append(approx_arrow[3])
                    else:
                        patch_corners.append(approx_arrow[0])
                        patch_corners.append(approx_arrow[1])
                        patch_corners.append(approx_arrow[3])
                        patch_corners.append(approx_arrow[2])


                    point_dist = []
                    point_dist.append(self.get_dist(patch_corners[0], patch_corners[1]))
                    point_dist.append(self.get_dist(patch_corners[1], patch_corners[2]))
                    point_dist.append(self.get_dist(patch_corners[2], patch_corners[3])) 
                    point_dist.append(self.get_dist(patch_corners[3], patch_corners[0]))

                    dist_error = 60 # threshold for error between edges of square

                    if abs(point_dist[0] - point_dist[1]) + abs(point_dist[1] - point_dist[2]) + abs(point_dist[2] - point_dist[3]) + abs(point_dist[3] - point_dist[0]) < dist_error:
                        check_sign = True
                    
                    if check_sign: # confirmation of square patch
                        cv.drawContours(output_mat.frame, contours, i, [255,0,0], 2)

                        for p in range(4):
                            x, y = patch_corners[p][0], patch_corners[p][1]
                            if p == 0:
                                cv.circle(output_mat.frame, (x, y), 1, (0,0,255), 3)
                            elif p == 1:
                                cv.circle(output_mat.frame, (x, y), 1, (0,255,0), 3)
                            elif p == 2:
                                cv.circle(output_mat.frame, (x, y), 1, (255,0,0), 3)
                        
                        sign_roi_corners = []
                        sign_src_corners = []

                        for j in range(4):
                            sign_src_corners.append(patch_corners[j])
                            sign_roi_corners.append([0.0, 0.0])

                        # clockwise order from top-left
                        sign_roi_corners[0][0] = 0.0
                        sign_roi_corners[0][1] = 0.0
                        sign_roi_corners[1][0] = 60.0
                        sign_roi_corners[1][1] = 0.0
                        sign_roi_corners[2][0] = 60.0
                        sign_roi_corners[2][1] = 60.0
                        sign_roi_corners[3][0] = 0.0
                        sign_roi_corners[3][1] = 60.0

                        sign_roi_corners = self.convert_to_numpy(sign_roi_corners).astype(np.float32)
                        sign_src_corners = self.convert_to_numpy(sign_src_corners).astype(np.float32)

                        # map the four corners to a perspective transform image
                        sign_perspective_transform = cv.getPerspectiveTransform(sign_src_corners, sign_roi_corners)
                        # warp perspective to correct the transformation
                        sign_roi = cv.warpPerspective(gray, sign_perspective_transform, (60,60), cv.INTER_LINEAR, cv.BORDER_REPLICATE, 0)
                        # get the binary output of the arrow patch using thresholding
                        sign_roi = cv.GaussianBlur(sign_roi, (5,5), 2, 2)
                        ret, sign_roi = cv.threshold(sign_roi, 127, 255, cv.THRESH_BINARY)

                        cv.imshow('Arrow', sign_roi)

                        # comparing the arrow's ROI to the template 60x60
                        check_arrow_left = 0 
                        check_arrow_right = 0
                        check_arrow_forward = 0

                        for w in range(60):
                            for h in range(60):
                                if sign_roi[h,w] == self.temp_left[h,w]:
                                    check_arrow_left += 1
                                else: 
                                    check_arrow_left -= 1

                                if sign_roi[h,w] == self.temp_right[h,w]:
                                    check_arrow_right += 1
                                else: 
                                    check_arrow_right -= 1

                                if sign_roi[h,w] == self.temp_forward[h,w]:
                                    check_arrow_forward += 1
                                else: 
                                    check_arrow_forward -= 1

                        # will consider all back pointing signs and forward, adjust the threshold or modify the conditions to fit your needs

                        if check_arrow_forward > 2600: #  all arrow pointing to front, use 3200 for FORWARD only
                            arrow_type = 'forward'
                        elif check_arrow_left > 2800:
                            arrow_type = 'left' 
                        elif check_arrow_right > 2800:
                            arrow_type = 'right'
                        elif check_arrow_left > 1900 and check_arrow_right > 2200: # condition for flip-left and flip-right
                            if check_arrow_right > 2400:
                                arrow_type = 'left'
                            else:
                                arrow_type = 'right'
                        else:
                            arrow_type = 'None'
                            # No arrow

        self.arrow.publish_msg(arrow_type)
        return output_mat


    def sort_by(self, element):
        return element[1]


    def order_points(self, pts):
        sorted_array = []

        for i in pts:
            sorted_array.append([i[0][0], i[0][1]])

        sorted_array.sort(key=self.sort_by)

        return sorted_array


    def convert_to_numpy(self, array):
        new_array = []
        for n in array:
            new_array.append(n)


        new_array = np.array(new_array)

        return new_array

