import cv2 as cv
import numpy as np
import math

from base.base_module import BaseModule
from base.configuration import Configuration
import base.opencv_gui as gui

from obstacle import Obstacle
from obstacle_mat import ObstacleMat

from mixed import Mixed


class ObstacleModule(Configuration):

    def __init__(self, base_module, configuration_directory, event, robot_size):
        super(ObstacleModule, self).__init__(
            configuration_directory=configuration_directory, event=event, module='obstacles', robot_size=robot_size)

        self.obstacles = []
        for obstacle in self.config['obstacles']:
            self.obstacles.append(
                Obstacle(configuration=obstacle, module=self))

        self.base = base_module
        base_module.add_debug_objects(self.obstacles)

        self.selected_obstacle = self.obstacles[0] if len(
            self.obstacles) > 0 else None
        self.colourspace_roi = None

        self.with_mixed = self.config['mixed']

        self.previous_circle_x = 0
        self.previous_circle_y = 0
        self.previous_circle_radius = 0
        self.possible_next_radius = (0,0)
        self.previous_min = 0
        self.avg_rad = 0
        self.k = 0
        self.count_missed = 0

    @staticmethod
    def is_enabled(base):
        return 'obstacles' in base.config['modules']

    def update(self):
        self.config['obstacles'][self.obstacles.index(
            self.base.selected_debug)] = self.base.selected_debug.export_configuration()
        self.save()

    def detect(self, input_mat, output_mat):
        fr = input_mat.clone()
        for obstacle in self.obstacles:
            # Extract obstacle from frame
            obstacle.mat = ObstacleMat(frame=fr.frame)
            obstacle.mat.extract_obstacle(thresh_lb=obstacle.lower_bound,
                                          thresh_ub=obstacle.upper_bound,
                                          kernel_size=obstacle.opening_closing_ksize,
                                          blur_ksize=obstacle.blur_ksize)

            # Find the obstacle using bounding rectangle, draw any if found
            positions, position_text, area, angle, box = obstacle.mat.bounding_rect(
                min_area=obstacle.min_area, max_area=obstacle.max_area)

            if box is not None:
                cv.drawContours(output_mat.frame, [box], 0, (0, 0, 0), 2)

            # Draw rectangle and publish ROS message
            x = -1
            y = -1
            if positions is not None:
                for tl_pt, br_pt in positions:
                    cv.rectangle(output_mat.frame, tl_pt, br_pt,
                                 obstacle.output_colour, thickness=2)

                if len(positions) != 0:
                    x = positions[0][1][0] - (positions[0][1][0] - positions[0][0][0]) / 2.
                    y = positions[0][1][1] - (positions[0][1][1] - positions[0][0][1]) / 2.

            obstacle.publish_msg(x, y, position_text, area, angle)

        if self.with_mixed:
            self.detect_mixed(fr, output_mat)

        return output_mat

    def detect_mixed(self, input_mat, output_mat):
        # The last two markers will make the Mixed Marker
        m1_pos = len(self.obstacles) - 2
        m2_pos = len(self.obstacles) - 1
        m1 = self.obstacles[m1_pos]
        m2 = self.obstacles[m2_pos]

        mixed_marker = Mixed(m1, m2, self)

        # why convert gray frame to brg frame??
        gray = cv.cvtColor(input_mat.frame, code=cv.COLOR_BGR2GRAY)

        #this debug object mat frame is not hsv
        #imshow('testing',hsv1)
        hsv1 = self.base.debug_objects[m1_pos].mat.frame
        hsv2 = self.base.debug_objects[m2_pos].mat.frame

        # hsv1 = cv.cvtColor(self.base.debug_objects[m1_pos].mat.frame, code=cv.COLOR_GRAY2BGR)
        # hsv2 = cv.cvtColor(self.base.debug_objects[m2_pos].mat.frame, code=cv.COLOR_GRAY2BGR)

        # Closing operation to fill the black pixels
        element = cv.getStructuringElement(cv.MORPH_ELLIPSE, (30, 30))
        hsv1 = cv.morphologyEx(hsv1, cv.MORPH_CLOSE, element)
        #hsv1 = cv.morphologyEx(hsv1, cv.MORPH_OPEN, element)
                # Closing operation to fill the black pixels
        element = cv.getStructuringElement(cv.MORPH_ELLIPSE, (30, 30))
        hsv2 = cv.morphologyEx(hsv2, cv.MORPH_CLOSE, element)
        
        #hsv1 = cv.cvtColor(hsv1, code=cv.COLOR_GRAY2BGR)
        #hsv2 = cv.cvtColor(hsv2, code=cv.COLOR_GRAY2BGR)
        #cv.imshow('hsv1',hsv1)

        hsv = np.bitwise_or(hsv1, hsv2)

        # Closing operation to fill the black pixels
        element = cv.getStructuringElement(cv.MORPH_ELLIPSE, (30, 30))
        hsv = cv.morphologyEx(hsv, cv.MORPH_CLOSE, element)

        canny = cv.Canny(hsv, 100, 255, 3)
        # Find all contours from canny edges
        im2, contours, hierarchy = cv.findContours(canny, cv.RETR_TREE, cv.CHAIN_APPROX_SIMPLE)

        # Get the moments

        mu = [None] * len(contours)

        for i in range(len(contours)):
            mu[i] = cv.moments(contours[i], False)

        # Get the mass centers

        mc = [None] * len(contours)

        x, y, area, angle = -1, -1, -1, -100


        for i in range(len(contours)-1):
            # area = mu[i]['m00']
            if mu[i]['m00'] > mixed_marker.min_area and mu[i]['m00'] < mixed_marker.max_area:
                mc[i] = (mu[i]['m10'] / mu[i]['m00'], mu[i]['m01'] / mu[i]['m00'])
                cv.drawContours(output_mat.frame, contours, i, [0, 165, 255], 2)
                area = mu[i]['m00']
                x, y = mc[i][0], mc[i][1]
                # print area
                # Find angle, useful?

                # angle, box = mixed_mat.find_angle(contours[i])

                # if box is not None:
                #     cv.drawContours(output_mat.frame, [box],0,(0,0,0),2)


        mixed_marker.publish_msg(x, y, 'none', area, angle)


        self.find_circle(hsv, mixed_marker, output_mat)


        cv.putText(hsv, text="""
        X: %d\tY:%d
        area: %d
        angle: %d
        """ %(x,y,area,angle)
                   , org=(10, 30),
                   fontFace=cv.FONT_HERSHEY_DUPLEX, fontScale=1, color=(0, 0, 255))
        

        cv.imshow("mixed", hsv)


    def find_circle(self, hsv, mixed_marker, output_mat):
        # Circle detect (not completely done yet):
        hsv = cv.cvtColor(hsv, code=cv.COLOR_GRAY2BGR)
        hsv = cv.cvtColor(hsv, code=cv.COLOR_BGR2GRAY)

        # element2 = cv.getStructuringElement(cv.MORPH_RECT, (10, 10))
        # hsv2 = cv.morphologyEx(hsv, cv.MORPH_CLOSE, element2)
        # canny2 = cv.Canny(hsv2, 100, 255, 3)


        minRadius = self.possible_next_radius[0]
        maxRadius = self.possible_next_radius[1]


        _,thresh = cv.threshold(hsv, 127, 255, 0)
        _, contours, _ = cv.findContours(thresh,cv.RETR_TREE,cv.CHAIN_APPROX_SIMPLE)

        circles = None
        max_contour_area = -1
        #search for the largest circle
        for i, c in enumerate(contours):
            if max_contour_area < cv.contourArea(contours[i]):
                contours_poly = cv.approxPolyDP(contours[i], 3, True)
                (cx, cy), cr = cv.minEnclosingCircle(contours_poly)
                circles = [int(cx), int(cy), int(cr)]




        #this is the old methon that we use
        # circles = cv.HoughCircles(hsv, cv.HOUGH_GRADIENT, 1, 20, \
        #     param1=50, param2=10, minRadius=minRadius, maxRadius=maxRadius)
        circle_x, circle_y, circle_area = -1, -1, -1
        
        if circles is not None:
            circles = np.uint16(np.around(circles))

            circle_x = circles[0]
            circle_y = circles[1]
            radius = circles[2]

            circle_area = math.pi * (radius ** 2)

            if circle_area > mixed_marker.min_area and circle_area < mixed_marker.max_area:
                p_gain = 0.01

                # y_error = y - circle_y
                # x_error = x - circle_x

                # x_error = circle_x - self.previous_circle_x
                # y_error = circle_y - self.previous_circle_y

                # dumb_radius = (area/math.pi) ** 0.5
                # print('dumb radius', dumb_radius)

                # radius_error = dumb_radius - radius
                # radius_error = radius - self.previous_circle_radius

                # circle_x += int(round(x_error * p_gain))
                # circle_y += int(round(y_error * p_gain))
                
                # if abs(radius_error) < radius:
                #     radius += radius_difference

                if self.k < 10:
                    max_rad = max(self.previous_circle_radius, radius)
                    min_rad = min(self.previous_circle_radius, radius)
                    self.previous_min = min_rad
                    radius = max_rad
                    radius_error = max_rad - min_rad
                    radius_difference = int(round(radius_error * p_gain))
                    self.k += 1
                else:
                    self.k = 0

                self.possible_next_radius = (radius - 2, radius + 2)

                self.previous_circle_y = circle_y
                self.previous_circle_x = circle_x
                self.previous_circle_radius = radius
                
                
                cv.circle(output_mat.frame, (circle_x, circle_y), 3, [0,255,0], -1, 3, 0)
                cv.circle(output_mat.frame, (circle_x, circle_y), radius, [0,0,255], 2, 3, 0)

            else:
                self.count_missed += 1
                circle_x, circle_y, circle_area = -1, -1, -1
        else:
            self.count_missed += 1
            circle_x, circle_y, circle_area = -1, -1, -1

        if self.count_missed > 3:
            # self.previous_circle_radiusq = 0
            avg = (self.previous_min + self.previous_circle_radius)/2
            self.possible_next_radius = (avg-10,avg+10)
            self.count_missed = 0

        mixed_marker.publish_msg(circle_x, circle_y, 'none', circle_area, -100, 'circle')

